/* wagon-ESP-OSC
 *  un wagon radio-guidé
 *  plateforme ESP8266 ESP-12
 *  commande moteur Olimex BB-L298
 *  protocole réseau OSC
 *  SNHack gepeto@du-libre.org
 *  Codes sous licences GPLV3
 */

/* projet LOCO-DOM
 *  IP du réseau 192.168.0.x/24
 *  les loco sont de x= 81 a 91
 *  ports 9081 a 9091
 * Connexions ESP-12 BBL-298
 * |-                |   ESP-12   |          |        -|
   |-                |      RST   |       TXD|        -|
   |LDR              | ADC        |       RXD|       - |
   |-                |CH_P        |    GPIO4 |  1 RFID |
   |4 BBL            |   GPIO16   |    GPIO5 |        -|
   |3 BBL            |   GPIO14   |    GPIO0 |        -|
   |led bleu2 BBL-298|   GPIO12   |    GPIO2 |        -|
   |led verte        |   GPIO13   |   GPIO15 |led rouge|
   |-                |   VCC 3V   |       GND|        -|
 *
 */

// WIFI
#include <ESP8266WiFi.h>
#include <WiFiUdp.h>
#include <Ticker.h> // gestion des timer
#define TAILLE_PUCE 12
#define TAILLE_TABLEAU_PUCES 480 // 40 x 12
//MDNSResponder mdns;
// OSC
#include <OSCBundle.h>
#include <OSCMessage.h>
#include <OSCBoards.h>
//port numbers for OSC
const unsigned int inPort = 9081;
//destination IP pour la trace OSC
IPAddress MonitIP(192, 168, 0, 99); // IP et port du moniteur
const unsigned int MonitPort = 9998; // suivi normal
IPAddress DebugIP(192, 168, 0, 123); // IP et port de debug
const unsigned int DebugPort = 9999;
WiFiUDP udpin;
WiFiUDP udpout;
Ticker zou; // OSC envoi position
Ticker init_rfid; // init port rfid
#include <ArduinoJson.h>
#include "FS.h"
// serial
#include <EspSoftSerialRx.h>
EspSoftSerialRx rfid;
const char* ssid     = "LOCOMO";
const char* password = "0123456789";
char nomhost[20]  = "LOCO-1";
// DHCP donc pas besoin
//IP
//IPAddress (192,168,IPC,IPD);
int IPC = 0;
int IPD = 81;
String nom = "LOCO-1";
char* OSCVar = "/LOCO-1";

// connect motor controller pins to Arduino digital pins
// ESP moteur traction
int enA = 12;
int in1 = 13;
int in2 = 14;
int Vitesse_actuelle = 0 ; // vitesse actuelle 
uint32_t temps_boucle_total = 14000 ; // en milli seconde
uint32_t temps_a_faire = 0;
uint32_t temps_depart = 0;
uint32_t temps_inter = 0;
uint32_t temps_tot_puce = 0;
int rampe_moyenne = 40;
char buf0[] = "00000000000000000000";
char tag[TAILLE_PUCE + 1];
unsigned char val = 0x00;
char TablePuce[TAILLE_TABLEAU_PUCES];
const int nbr_max_de_puce = 20;
const int distance_entre_puce = 43 ; // cm
const int distance_totale = 516 ; //cm
int nbr_de_puce = 0;
int no_puce_actuelle = 0;
int puce_a_trouver = 0;
int puce_derniere_lue = 0;
// tableau des puces rfid 20x12
//char TableauPuces[nbr_max_de_puce][TAILLE_PUCE] = {
char TableauPuces[nbr_max_de_puce][TAILLE_PUCE + 1] = {
  "999999999999", // puce 0
  "0103FA188262", // 1
  "0103FA186989", // 2
  "0103FA187292", // 3
  "0103FA19E908", // 4
  "0103FA19FA1B", // 5
  "0103FA199475", // 6
  "0103FA1A6F8D", // 7
  "0103FA1933D2", // 8
  "0103FA1AE604", // 9
  "0103FA197A9B", // 10
  "0103FA1A09EB", // 11
  "0103FA19DE3F", // 12
  "000000000000", // 13
  "000000000000", // 14
  "000000000000", // 15
  "000000000000", // 16
  "000000000000", // 17
  "000000000000", // 18
  "000000000000", // 19
};
uint32_t TableauTemps[nbr_max_de_puce];


static int32_t sequencenumber = 0;
int commande = 0; // no de la commande en cours

void setup()
{
  Serial.begin(115200);
  // set all the motor control pins to outputs
  pinMode(enA, OUTPUT);
  pinMode(in1, OUTPUT);
  pinMode(in2, OUTPUT);
  // serial RFID
  rfid.begin(9600, 2); // sortie GPIO2
  Serial.println(ESP.getChipId(), HEX);
  //setup ethernet part
  Serial.println();
  Serial.print("Connecting to ");
  Serial.println(ssid);
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  //  DHCP donc pas fixé ici
  //WiFi.config(IPAddress(192, 168, IPC, IPD), IPAddress(192, 168, IPC, IPD), IPAddress(255, 255, 255, 0));
  Serial.println("");
  Serial.println("WiFi connected");
  Serial.println(ssid);
  Serial.println("IP address: ");
  Serial.println(WiFi.localIP());

  Serial.println("host: ");
  Serial.println(WiFi.hostname());
  WiFi.hostname(nomhost);
  Serial.println(WiFi.hostname());
  Serial.println("Starting UDP");
  udpin.begin(inPort);
  Serial.print("Local port: ");
  Serial.println(udpin.localPort());

  delay(100);
  // envoi le bonjour
  MonitMsg("BONJOUR", (int32_t)ESP.getChipId());
  //recupere le s puce du fichier /puces.buf
  if (!SPIFFS.begin()) {
    Serial.println("Failed to mount file system");
  }
  File configFile = SPIFFS.open("/puces.buf", "r");
  if (!configFile) {
    Serial.println("pb de lecture de /puces.buf");
    nbr_de_puce = nbr_max_de_puce;
    MonitMsg("nbr max de puce ", nbr_de_puce);
  } else {
    char *erg = "";
    configFile.readBytes(erg, 1);// nbr_de_puce dans le premier enregistrement du fichier
    nbr_de_puce = (int) erg;
    MonitMsg("fichier config utilise ", nbr_de_puce);
    for (int i = 1; i <= nbr_de_puce; i++) {  // init tableau pour chaque puce
      configFile.readBytes( TableauPuces[i] , TAILLE_PUCE );
      MonitMsg(TableauPuces[i] , i);
    }
    configFile.close();
  }
  // placer jesuisla en boucle timer 1s
  //zou.attach(5, jesuisla);
  // init_rfid.attach(10, rfid_port_init); ?
  // En Avant;
  digitalWrite(in1, HIGH);
  digitalWrite(in2, LOW);
  //Vitesse_moteur(1000);
}
void rfid_port_init() { // re initialisation du port rfid les 10s
  Serial.println("init rfid");
  rfid.service();
}
void jesuisla() { // envoie OSC /msg if position vitesse
  OSCMessage zouzou(OSCVar);
  zouzou.add("cmd puce vits tps ");
  zouzou.add((int32_t)commande); // commande en cours
  zouzou.add((int32_t)no_puce_actuelle);
  zouzou.add((int32_t)Vitesse_actuelle);
  zouzou.add((int32_t)(millis() / 1000));
  udpout.beginPacket(MonitIP, MonitPort);
  zouzou.send(udpout);
  udpout.endPacket();
  udpout.beginPacket(DebugIP, DebugPort);
  zouzou.send(udpout);
  udpout.endPacket();
  zouzou.empty(); // free space occupied by
}
void Arriere() {
  digitalWrite(in1, LOW);
  digitalWrite(in2, HIGH);
}
void Avant() {
  digitalWrite(in1, HIGH);
  digitalWrite(in2, LOW);
}
int Lit_Serie() {
  strncpy(tag, buf0,  TAILLE_PUCE + 1);
  for (int i = 0; i <= (TAILLE_PUCE - 1 ); i++) {
    val = 0x00;
    if (rfid.read(val)) {
      // on saute avec le start 0x02 et stop bit 0x03
      if ((val == 0x03) || (val == 0x02)) {
        break;
      }
      //Serial.println(val, HEX );
      tag[i] = (char)val;
      if (i == TAILLE_PUCE - 1) { // stop bit complete
        tag[TAILLE_PUCE] = '\0'; // on ferme ici
        //        Serial.print("Tag read: ");
        MonitMsg(tag, 1);
        Serial.print("tag ");
        Serial.println(tag);
        return (0);
      }
    }
  }
  return (1);
}
int Lit_rfid() {
  // mise en forme de la lecture rfid
  //retourne le no_puce
  //Serial.println("attente rfid");
  int i = 0;
  char tag[TAILLE_PUCE + 1];
  strncpy( tag, buf0, TAILLE_PUCE) + 1;
  for (i = 0; i <= (TAILLE_PUCE - 1); i++) {
    val = 0x00;
    if (rfid.read(val)) {
      tag[i] = '\0'; // Null
      // on saute avec le start 0x02 et stop bit 0x03
      if ((val == 0x0D) || (val == 0x0A) || (val == 0x03) || (val == 0x02)) {
        break;
      }
      //      Serial.print(i);
      //      Serial.print(" ");
      //      Serial.print((char)val);
      //      Serial.print(" ");
      //    Serial.println(val, HEX );
      tag[i] = (char)val;

      if (i == (TAILLE_PUCE - 1) ) { // 11 digit read is complete
        tag[TAILLE_PUCE] = '\0'; // Null on ferme la chaine
        //Serial.print(tag);
        puce_derniere_lue = CherchePuce( tag );
        // Only fire the strike plate if this tag was found in the database
        if ( puce_derniere_lue != 99 )
        {
          //Serial.print(" lirfid puce ");
          //Serial.println(puce_derniere_lue);
          MonitMsg("puce ", puce_derniere_lue);
          return (puce_derniere_lue);
        } else {
          MonitMsg(tag, 99);
        }
      }
    }
  }
}

int CherchePuce( char tag[TAILLE_PUCE] ) {
  for (int puce = 0; puce < nbr_de_puce; puce++) {
    // Check if the tag value matches this row in the tag database
    if (strcmp(tag, TableauPuces[puce]) == 0)
      return (puce);
  }
  // Si pas trouvé on retourne 99
  return (99);
}


bool MonitMsg(char debugbla[], int debugint) {
  OSCMessage envoi(OSCVar);
  envoi.add(debugbla);
  envoi.add(debugint);

  udpout.beginPacket(MonitIP, MonitPort);
  envoi.send(udpout);
  udpout.endPacket();

  udpout.beginPacket(DebugIP, DebugPort);
  envoi.send(udpout);
  udpout.endPacket();
  envoi.empty();
}
int Vitesse_moteur(int moteur, int accel) {
  int i = 0;
  moteur = abs(moteur);
  if (moteur > 1024) // filtre le max
    moteur = 1024;

  MonitMsg("vitesse", moteur);
  //Serial.println(moteur);
  //Serial.println(Vitesse_actuelle);
  if ( accel < -500 ) accel = rampe_moyenne;
  if ( accel < 0 )    moteur = -moteur; // ARR
  if ( accel > 500 ) accel = rampe_moyenne;
  // sinon on gare accel donne
  accel = abs(accel);
  if (Vitesse_actuelle > moteur) {
    for (i = Vitesse_actuelle; i >= moteur; i = i - 20) {
      if ( i >= 0 ) {
        //Avant;
        digitalWrite(in1, HIGH);
        digitalWrite(in2, LOW);
      } else {
        //Arriere;
        digitalWrite(in1, LOW);
        digitalWrite(in2, HIGH);
      }
      analogWrite(enA, abs(i));
      delay(accel); // rampe
    }
  } else {
    for ( i = Vitesse_actuelle; i <= moteur; i = i + 20) {
      if ( i >= 0 ) {
        //Avant;
        digitalWrite(in1, HIGH);
        digitalWrite(in2, LOW);
      } else {
        //Arriere;
        digitalWrite(in1, LOW);
        digitalWrite(in2, HIGH);
      }
      analogWrite(enA, abs(i));
      delay(accel); // rampe
    }
  }
  Vitesse_actuelle = moteur;
  analogWrite(enA, abs(moteur));
  //no_puce_actuelle = 97;
}
// COMMANDES
void ArretMoteur(OSCMessage &msg) {
  commande = 0;
  Serial.println("ArretMoteur /stop");
  MonitMsg("STOP", 0);
  analogWrite(enA, 0);
}
void MoteurControl(OSCMessage &msg) {
  // met le moteur à la vitesse reçue, accel
  commande = 1;
  Serial.println("/moteur");
  Vitesse_moteur(msg.getInt(0), msg.getInt(1));
}

void InitTableauPuces(OSCMessage &msg) {
  int i = 0;
  int puce = 0;
  char *buf[TAILLE_PUCE];
  Serial.println("/inittableaupuce");
  commande = 7;
  if (msg.getInt(0) > 0) {
    nbr_de_puce = (int) msg.getInt(0);
  } else {
    nbr_de_puce = nbr_max_de_puce;
  }
  Serial.println(nbr_de_puce);
  MonitMsg("INIT Tableau", nbr_de_puce);
  // tour pour nbr de puce total
  MonitMsg("Apprendre puce 1", 0 );
  strncpy( tag, "",  TAILLE_PUCE);
  while (Lit_Serie() != 0) {
    Serial.println("puce ? 1 ");
    delay(2000);
  }
  Serial.println(tag);
  strncpy( TableauPuces[1], tag,   TAILLE_PUCE);
  MonitMsg(TableauPuces[1], 1);
  MonitMsg("Apprendre puces", nbr_de_puce );
  for (puce = 2; puce <= nbr_de_puce; puce++) { // On doit déjà connaitre la puce 1
    strncpy(tag, "",   TAILLE_PUCE);
    while (Lit_Serie() != 0) {
      Serial.print("puce? ");
      Serial.print(puce);
      delay(2000);
      Serial.println(" :");
    }
    if (strcmp(tag, TableauPuces[1]) == 0 ) {
      nbr_de_puce = puce;
      break;
    }
    strncpy( TableauPuces[puce], tag,  TAILLE_PUCE) ;
    Serial.print(tag);
    Serial.print(" ");
    Serial.println(puce);
    MonitMsg(tag, puce );
  }
  Serial.print  ("nbr de puce ok ");
  Serial.println(puce);
  MonitMsg("nbr de puce ok", puce);
}
void SauverTableauPuces(OSCMessage &msg) {
  uint8_t *buf[TAILLE_PUCE];

  MonitMsg("sauvegarde des puces", nbr_de_puce);
  TableauPuces[0][0] = (char)nbr_de_puce ;
  File configFile = SPIFFS.open("/puces.buf", "w");
  if (!configFile) {
    Serial.println("pb d'ecriture de /puces.buf");
  } else {
    for (int i = 1; i <= nbr_de_puce; i++) {
      //      strncpy(buf, (uint8_t)TableauPuces[i], TAILLE_PUCE);
      //      configFile.write(buf  , TAILLE_PUCE) ;
      //      MonitMsg(tag, i);
    }
  }
}
void Initialisation(OSCMessage &msg) {
  // la loco fait un tour complet doucement pour connaitre le nombre de puce
  // la premiere etant la puce 1
  // ou on lui mpresente les puces une par une
  // puis la loco fait un tour complet pour connaitre le temps nécessaire
  // et vérifier la lecture des puces RFID , voir faire l'init du tableau?
  // elle ne connait au départ que la puce 0, position 0
  // on ecrit le tableau sur le fs
  int i = 0;
  int puce = 0;
  commande = 2;
  val = 0x00;
  zou.detach();
  strncpy( tag, buf0, TAILLE_PUCE + 1);
  if (msg.getInt(0) > 0) {
    nbr_de_puce = (int) msg.getInt(0);
  } else {
    nbr_de_puce = nbr_max_de_puce;
  }
  Serial.print("/init");
  Serial.println(nbr_max_de_puce);
  MonitMsg("INIT", nbr_max_de_puce);

  Vitesse_moteur(0, 0);
  MonitMsg("Apprendre puce 1", 1 );
  Vitesse_moteur(650, 0);
  while (Lit_Serie() == 0)
    delay(500); // les puces sont eloignes de plus de 1/2 s

  strncpy( TableauPuces[1], tag,  TAILLE_PUCE) ; // premiere puce 1
  MonitMsg("puce 1 ", 0);
  MonitMsg(TableauPuces[1], 0);
  for (puce = 2; puce <= nbr_de_puce; puce++) { // On doit déjà connaitre la puce 1
    if (strcmp(TableauPuces[1], TableauPuces[puce])) {
      nbr_de_puce = puce;
      Vitesse_moteur(0, 0);
      MonitMsg("nbr de puce", puce);
      break;
    }
    while (Lit_Serie() == 0)
      delay(500);
    strncpy( TableauPuces[puce], tag,  TAILLE_PUCE) ;
  }
  Aller(1, -650, 0); //retourner sur la puce 1 et y stopper
  uint32_t temps = (uint32_t)millis();
  //rouler a fond
  temps_depart = (uint32_t)millis();
  Aller(1, 1000, 0);
  for (puce = 2; puce <= nbr_de_puce ; puce++) {
    if (strcmp(TableauPuces[1], TableauPuces[puce])) { // puce 1 re-trouve
      break;
    }
    while (Lit_Serie() == 0)
      delay(100);
    TableauTemps[puce] = (uint32_t)millis() - temps;
    MonitMsg(tag, puce);
  }
  temps_boucle_total = (uint32_t)millis() - temps - 800; // gaff au temp de lecture
  MonitMsg("temps boucle", temps_boucle_total );
  TableauTemps[0] = temps_boucle_total;
  // enregistrer les parametres
  commande = 0;
  Aller(1, -650, 0); //retourner sur la puce 1 et y stopper
  //zou.attach(5, jesuisla);
}
void Rendezvous(OSCMessage & msg) {      // A FINIR
  // /goto iii puce temps rampe : se rendre a la puce xx en tant de temps
  commande = 3;
  int rampe = 0;
  int distance = 0;
  float vitesse = 0;
  int puce = 0;
  puce_a_trouver = (int) msg.getInt(0);
  temps_a_faire = (uint32_t)msg.getInt(1) ;
  rampe = (int) msg.getInt(2); // si <0 alors marche arriere
  if (puce_derniere_lue == 0  ||  puce_derniere_lue > 90  ) { // on ne connait pas la puce depart
    Serial.print("/rdv de ");
    Serial.println(puce_derniere_lue);
    Vitesse_moteur(900, -1);
    while (Lit_rfid() >= 90 || puce_derniere_lue == 0  ) // on cherche la prochaine, renseigne puce_derniere_lue
      delay(200);
  }
  no_puce_actuelle = puce_derniere_lue; // ?
  Vitesse_moteur(0, -1);
  //Serial.print("/rdv de ");
  //Serial.print(puce_derniere_lue);
  //Serial.print(" a trouver ");
  //Serial.println(puce_a_trouver);
  temps_depart = (uint32_t)millis();
  // 1 vitesse constante sur le trajet sauf accel
  // v= vitess max / (Dtot/Ttot)*t/d
  // puce devant plus grande et rampe poru marche avant (>-1)
  if ((puce_a_trouver >= puce_derniere_lue) && rampe >= 0) { //si meme puce alors un tour complet
    distance = (puce_a_trouver - puce_derniere_lue) * distance_entre_puce;
  } else {
    if ((puce_a_trouver < puce_derniere_lue) && rampe >= 0) {
      // puce devant plus petite et marche avant
      distance = (puce_a_trouver - puce_derniere_lue + nbr_de_puce ) * distance_entre_puce;
    } else {
      if ((puce_a_trouver >= puce_derniere_lue) && rampe < 0) { //arriere
        distance = (nbr_de_puce - puce_a_trouver - puce_derniere_lue) * distance_entre_puce;
      } else {
        if ((puce_a_trouver < puce_derniere_lue) && rampe < 0) { //arriere
          distance = (puce_derniere_lue - puce_a_trouver) * distance_entre_puce;
        }
      }
    }
  }
  if (distance == 0) distance = distance_totale;
  vitesse =  ((float)distance_totale / (float)temps_boucle_total) * (((float)temps_a_faire / (float)distance);
  vitesse = 1000 / vitesse;
  if (vitesse > 1010) MonitMsg("ERR temps trop court tps mini", temps_boucle_total );
  Serial.println(vitesse);
  MonitMsg("vitesse calculee", (int) vitesse);
  //Aller(puce_a_trouver, (int)vitesse, rampe);
  temps_depart = (uint32_t)millis();
  Vitesse_moteur((int)vitesse, rampe); // on roule , pendant un certain temps
}
void SeRendre(OSCMessage & msg) {
  // /aller ii puce vitesse : se rendre a la puce xx a n vitesse [acceleration]
  commande = 4;
  Aller((int)msg.getInt(0), (int)msg.getInt(1), (int)msg.getInt(2) );
}
void LitPuce(OSCMessage & msg) {
  commande = 6;
  Serial.println("lit");
  zou.detach();
  MonitMsg("lit", 9000);
}
void Ou(OSCMessage & msg) {
  commande = 9;
  Serial.println("ou");
  zou.attach(5, jesuisla);
  MonitMsg("ou", 9000);
}
bool Aller(int no_puce, int vitesse,  int rampe) {
  // aller jusqu'a la puce no_puce a vitesse
  Serial.println("Aller");
  MonitMsg("aller", no_puce);
  puce_a_trouver = no_puce;
  temps_depart = (uint32_t)millis();
  temps_tot_puce = temps_depart;
  Vitesse_moteur(vitesse, rampe);
  return (true);
}
void loop()
{
  // Lire la comande et executer le debut
  OSCMessage msgIN;
  int size;
  if ( (size = udpin.parsePacket()) > 0)
  {
    while (size--)
      msgIN.fill(udpin.read());

    if (!msgIN.hasError()) {
      // /stop immediat moteur et commande
      msgIN.dispatch("/stop", ArretMoteur);  // commande=0

      // /moteur ii vitesse +/- rampe (av/arr)
      //        rampe  <0 : marche arriere, 0 brutal
      msgIN.dispatch("/moteur", MoteurControl); //  commande=1

      // /init fait un tour pour decouvrir les puces, le parcours distance_totale et temps_boucle_totale
      msgIN.dispatch("/init", Initialisation);  // commande=2

      // /goto iii puce temps(ms) +/-rampe (av/arr) va à la puce dans le temps donné
      msgIN.dispatch("/goto", Rendezvous);      // commande=3 

      // /aller iii puce vitesse rampe
      msgIN.dispatch("/aller", SeRendre);       // commande=4

      // /lit renvoi le code de la puce lue
      msgIN.dispatch("/lit", LitPuce);          //   commande=6

      // /puces rempli le tableau avec les puce presentees a larret
      msgIN.dispatch("/puces", InitTableauPuces);  // commande=7

      // /sauvepuces enregistre le tableau dans la flash
      msgIN.dispatch("/sauvepuces", SauverTableauPuces); // commande=8

      // /ou renvoi la derniere puce lu
      msgIN.dispatch("/ou", Ou);                //  commande=9
    }
    udpin.flush();
  }
  // Lire une puce et continuer une commande en cours
  if (commande != 0) {
    if (commande != 2 && commande != 7) {
      if (commande == 6) {
        Lit_Serie();
      } else {
        puce_derniere_lue = Lit_rfid();
        if (puce_derniere_lue > 0) {
          temps_inter = (uint32_t)millis() - temps_depart ;
          temps_tot_puce += temps_inter;
          MonitMsg("temps ", temps_inter );
          //Serial.print("puce lue ");
          //Serial.print(puce_derniere_lue);
          //Serial.print(" a trouver ");
          //Serial.println(puce_a_trouver);
          if ((commande == 3 && Vitesse_actuelle > 800){
            if( temps_a_faire - ((uint32_t)millis() - temps_depart)) < 1500 ){ // ralentir
            Vitesse_actuelle = Vitesse_actuelle - 100
            if (Vitesse_actuelle<0) Vitesse_actuelle=0;
            analogWrite(enA, Vitesse_actuelle);
            }
          }
          if ((commande == 4 || commande == 3) && puce_derniere_lue == puce_a_trouver) {
            //stopmoteur
            Vitesse_moteur(0, -1); //rampe a 0
            //Serial.println(" ok ");
            MonitMsg("puce trouve" , puce_derniere_lue);
            MonitMsg("en ", temps_tot_puce - temps_depart );
            temps_tot_puce = 0;
            no_puce_actuelle = puce_a_trouver;
            //puce_derniere_lue = 98;
            puce_a_trouver = 98;
            commande = 0;
          }
          no_puce_actuelle = puce_derniere_lue;
        }
      }
    }
    delay(30);
  }

}
